"""
Simple MongoDB Document Wrapper

Adds useful functionality beyond standard dict type.

"""

import re
import unicodedata

from pymongo.database import Database
from bson.objectid import ObjectId, InvalidId

_slugify_strip_re = re.compile(r'[^\w\s-]')
_slugify_hyphenate_re = re.compile(r'[-\s]+')


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    From Django's "django/template/defaultfilters.py".
    """
    if not isinstance(value, unicode):
        value = unicode(value)
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    value = unicode(_slugify_strip_re.sub('', value).strip().lower())
    return _slugify_hyphenate_re.sub('-', value)


class NoDatabaseSet(Exception):
    """ No database on document """
    pass


class DocumentNotSaved(Exception):
    """ Document not saved """
    pass


class CollectionProxy(type):
    """
    Meta class for Document, proxy unfound attributes to pymongo collection

    """

    def __init__(cls, *args, **kwargs):
        """
        On creating a Document-type class, attach the pymongo collection to the
        collection attribute.

        """
        if cls.__database__:
            if not isinstance(cls.__database__, Database):
                raise NoDatabaseSet()
            cls.collection = cls.__database__[cls.__collection__ or
                                              cls.__name__]
        return super(CollectionProxy, cls).__init__(*args, **kwargs)

    def __getattr__(cls, attr):
        """
        Effectively proxy methods on the pymongo collection by catching
        unmatched attributes of the Document class.

        """
        if not hasattr(cls, 'collection'):
            raise(AttributeError)
        return getattr(cls.collection, attr)


class Document(dict):
    """
    SimpleMongo Document/Collection wrapper

    """

    __metaclass__ = CollectionProxy

    # if unset, uses class name for Mongo collection
    __collection__ = None
    __database__ = None

    """ Class methods """

    @classmethod
    def find(cls, *args, **kwargs):
        if not hasattr(cls, 'collection'):
            raise NoDatabaseSet()
        return cls.collection.find(*args, as_class=cls, **kwargs)

    @classmethod
    def find_one(cls, *args, **kwargs):
        if not hasattr(cls, 'collection'):
            raise NoDatabaseSet()
        return cls.collection.find_one(
            *args, as_class=cls, **kwargs)

    @classmethod
    def create_indexes(cls):
        """ create indexes if they don't already exists """
        if not hasattr(cls, 'collection'):
            raise NoDatabaseSet()
        if hasattr(cls, 'index_by') and type(cls.index_by) is list:
            for index in cls.index_by:
                if type(index) is dict and index.get('fields'):
                    options = index.get('options', {})
                    if not 'background' in options:
                        options['background'] = True
                    cls.collection.ensure_index(
                        index['fields'], **options)

    @classmethod
    def by_id(cls, id_value, *args, **kwargs):
        """ lookup record by id """
        if not isinstance(id_value, ObjectId):
            try:
                id_value = ObjectId(id_value)
            except InvalidId:
                pass
        return cls.find_one({'_id': id_value}, *args, **kwargs)

    @classmethod
    def by_slug(cls, slug, *args, **kwargs):
        """ lookup record by slug """
        return cls.find_one({'slug': slug}, *args, **kwargs)

    """ Instance methods """

    def __init__(self, *args, **kwargs):
        if not self.__database__:
            raise(NoDatabaseSet)
        self.update = self.__instance_update
        super(Document, self).__init__(self, *args, **kwargs)

    def save(self):
        return self.collection.save(self)

    def __instance_update(self, *args, **kwargs):
        if not self.get('_id'):
            raise DocumentNotSaved()
        if not args:
            result = self.collection.update({'_id': self['_id']},
                                            self, **kwargs)
        else:
            result = self.collection.update({'_id': self['_id']},
                                            *args, **kwargs)
        if not 'reload_after' in kwargs or kwargs['reload_after']:
            self.reload()
        return result

    def reload(self):
        if not self.get('_id'):
            raise DocumentNotSaved()
        super(Document, self).update(self.__class__.by_id(self['_id']))

    def delete(self):
        if not self.get('_id'):
            raise DocumentNotSaved()
        return self.collection.remove({'_id': self['_id']})

    def slugify(self, from_fields=None, from_text=None):
        """ Generate slug from specified fields or text """
        if not from_text:
            if not from_fields and not hasattr(self, 'slug_fields'):
                raise ValueError('No implicit or explicit fields to slugify')
            in_fields = from_fields or self.slug_fields
            if type(in_fields) is list:
                from_text = ' '.join(
                    [unicode(self[data]) for data in in_fields])
            else:
                from_text = unicode(self[in_fields])
        slug = slugify(from_text)
        next_id = self.__database__['__slug_counter'].find_and_modify(
            query={'_id': slug},
            update={'$inc': {'count': 1}},
            upsert=True,
            new=True)
        if next_id.get('count') == 1:
            return slug
        return '%s-%d' % (slug, next_id.get('count'))

    @property
    def id(self):
        """ id property returns mongodb's object _id """
        return self.get('_id')
